#include "MoveGeneration.h"
#include "ChessGame.h"

using namespace chessy;

namespace chessy
{
    template class PawnPushGenerator<WHITE_SIDE>;
    template class PawnPushGenerator<BLACK_SIDE>;

    template class PawnDoublePushGenerator<WHITE_SIDE>;
    template class PawnDoublePushGenerator<BLACK_SIDE>;

    template class PawnCaptureGenerator<WHITE_SIDE, LEFT>;
    template class PawnCaptureGenerator<WHITE_SIDE, RIGHT>;
    template class PawnCaptureGenerator<BLACK_SIDE, LEFT>;
    template class PawnCaptureGenerator<BLACK_SIDE, RIGHT>;
}


template<Side side>
typename PawnPushGenerator<side>::MoveIterator PawnPushGenerator<side>::begin(void) const
{
    const Board& board = chessGame.getBoard();
    Bitboard movedPieces = board.getPawns<side>();
    if (side == WHITE_SIDE)
        movedPieces.moveNorthOne();
    else
        movedPieces.moveSouthOne();
    
    movedPieces &= ~board.getOccupied();
    //movedPieces &= Bitboard(~0xFF000000000000FFULL);

    return MoveIterator{ movedPieces };
}


template<Side side>
typename PawnPushGenerator<side>::MoveIterator PawnPushGenerator<side>::end(void) const
{
    return MoveIterator{ 0 };
}


template<Side side>
typename PawnDoublePushGenerator<side>::MoveIterator
    PawnDoublePushGenerator<side>::begin(void) const
{
    const Board& board = chessGame.getBoard();
    Bitboard movedPieces = side == WHITE_SIDE ?
        board.getWhitePawns() & Bitboard(0xFFULL << 8) :
        board.getBlackPawns() & Bitboard(0xFFULL << 48);
    if (side == WHITE_SIDE)
        movedPieces.moveNorth(2);
    else
        movedPieces.moveSouth(2);
    
    movedPieces &= ~board.getOccupied();

    return MoveIterator{ movedPieces };
}


template<Side side>
typename PawnDoublePushGenerator<side>::MoveIterator
    PawnDoublePushGenerator<side>::end(void) const
{
    return MoveIterator{ 0 };
}


template<Side side, Leftright leftright>
typename PawnCaptureGenerator<side, leftright>::MoveIterator
    PawnCaptureGenerator<side, leftright>::begin(void) const
{
    const Board& board = chessGame.getBoard();
    Bitboard movedPieces = side == WHITE_SIDE ?
        board.getWhitePawns() :
        board.getBlackPawns();
    if (side == WHITE_SIDE)
        movedPieces.moveNorthOne();
    else
        movedPieces.moveSouthOne();
    
    if (leftright == LEFT)
        movedPieces.moveEastOne();
    else
        movedPieces.moveWestOne();
    
    movedPieces &= (side == BLACK_SIDE ? board.getWhites() : board.getBlacks());
    //movedPieces &= Bitboard(~0xFF000000000000FFULL);

    return MoveIterator{ movedPieces };
}


template<Side side, Leftright leftright>
typename PawnCaptureGenerator<side, leftright>::MoveIterator
    PawnCaptureGenerator<side, leftright>::end(void) const
{
    return MoveIterator{ 0 };
}

const std::array<Bitboard, 64> KnightMoveGenerator::moveSets = generateKnightMoves();

Bitboard KnightMoveGenerator::generateFromIndex(Index i) {
    int row = i.getRow();
    int column = i.getColumn();
    Bitboard result = 0;
    auto makeBitboardIfValid = [](int row, int column) -> Bitboard {
        return row < 8 && column < 8 && row >= 0 && column >= 0 ?
            Bitboard::fromIndex({row, column}) : Bitboard{ 0 };
    };
    result |= makeBitboardIfValid(row - 2, column - 1);
    result |= makeBitboardIfValid(row - 1, column - 2);
    result |= makeBitboardIfValid(row + 2, column - 1);
    result |= makeBitboardIfValid(row + 1, column - 2);
    result |= makeBitboardIfValid(row - 2, column + 1);
    result |= makeBitboardIfValid(row - 1, column + 2);
    result |= makeBitboardIfValid(row + 2, column + 1);
    result |= makeBitboardIfValid(row + 1, column + 2);
    return result;
}


std::array<Bitboard, 64> KnightMoveGenerator::generateKnightMoves(void)
{
    std::array<Bitboard, 64> moves{};
    for (int8_t i = 0; i < moves.size(); i++) {
        moves[i] = generateFromIndex(Index{ i });
    }
    return moves;
}


Bitboard PrimitiveQueenMoveGenerator::generateAttackSet(Index i,
    Bitboard enemies, Bitboard friendly, bool straight, bool diagonal)
{
    Bitboard queen = Bitboard::fromIndex(i);
    Bitboard occupied = (enemies | friendly) & ~queen;

    // strange c++ incoming
    auto iterate = [queen, occupied] (void (Bitboard::*moveStep)()) -> Bitboard {
        Bitboard result = 0;
        int a = 0;
        for (Bitboard r = queen; !(r & occupied) && r;) {
            (r.*moveStep)();
            result |= r;
            a++;
        }
        return result;
    };

    Bitboard possibleMoves = 0;
    if (straight)
        possibleMoves |=
            iterate(&Bitboard::moveNorthOne) |
            iterate(&Bitboard::moveSouthOne) |
            iterate(&Bitboard::moveEastOne) |
            iterate(&Bitboard::moveWestOne);
    if (diagonal)
        possibleMoves |=
            iterate(&Bitboard::moveNEOne) |
            iterate(&Bitboard::moveSEOne) |
            iterate(&Bitboard::moveNWOne) |
            iterate(&Bitboard::moveSWOne);   
    return (possibleMoves & ~friendly);
}


Bitboard KingMoveGenerator::generateKingMoves(Bitboard king)
{
    return king.northOne() | king.southOne() | king.eastOne() | king.westOne() |
        king.neOne() | king.nwOne() | king.swOne() | king.seOne();
}

